%{
#include "togl.h"

#include "list.h"
#include "routine.h"
#include "track.h"
#include "movie.h"
#include "body.h"	// for track enum

#ifdef WIN32
#include <malloc.h>
#define alloca _alloca
#else
#include <stdlib.h>
#endif



#define DOERROR(msg) {yyerror(msg); YYERROR;}

// Comes from the parser
extern int yylineno;
extern FILE *yyin;
const char *yyfilename;
extern int yylex(void);
extern Tcl_Interp *currentInterp;

// Local prototypes
void yyerror(const char *msg);

// Local vars
class Routine *routine;
int parserMode;

static class Key *key;
static class Figure *figure;
static class Movie *movie;
static class Modif modif;
static class List<Bezier3D> *path;
static class List<Chunk> *chunks;
static class List<Modif> *modifs;
static class List<Note> *notes;
static double doublelist[10];
static double v1, v2;
%}

//------------------------------------------------------------------------------------------------

%union {
	int integer;
	double numeric;
	char *string;
	struct {int start, end;} time;
	struct {int time; double value;} point;
	struct {int time; double x,y,z;} point3D;
	struct {int start, end; char *string;} annotation;
	struct {double *v;} doublelist;
	struct {char *name, *type; int sex;} dancerinfo;
	struct {void *path;} path;		// class List<Bezier3D> *path;
}

%token <string> tSTRING tNAME tDESCRIPTION tCOPYRIGHT tTEXT tFILENAME
%token tDANCERS tINIT tKEY tROUTINE tMOVIE tNOTES tMAP tWITH tREJECT tCOLOR tINCLUDENOTES
%token tBEZIER tHANDLES tHANDLELEFT tHANDLERIGHT tLINEAR
%token tEYE tCENTER
%token tMAN tWOMAN
%token <integer> tINTEGER tBEATS tRHYTHM tBPM tTRACKNAME tJOINTNAME
%token <numeric> tDOUBLE

%type <dancerinfo> dancerInfo
%type <path> path
%type <time> time
%type <point> point chunks
%type <point3D> point3D path_entries
%type <integer> dancer sex
%type <doublelist> bezier bezier3D
%type <string> name
%type <annotation> annotation notes
%type <numeric> number

%%

/************************************************************************************************/
/*              Top most      						                                            */
/************************************************************************************************/

input		:	input statement
			|	/* empty */
			;

statement	:	tBEATS '=' tINTEGER			{ routine->nbBeats = $3; }

			|	tDANCERS '=' 
				{if (routine->nbDancers>0) DOERROR("dancers already defined");} dancers

			|	tRHYTHM '=' tINTEGER		{ routine->rhythm = $3; }

			|	tBPM '=' tINTEGER			{ routine->bpm = $3; }

			|	tNAME '=' tSTRING			{ routine->name = $3; }

			|	tDESCRIPTION '=' tSTRING 	{ routine->description = $3; }

			|	tCOPYRIGHT '=' tSTRING 		{ routine->copyright = $3; }

			|	tTEXT '=' tSTRING 			{ routine->text = $3; }

			|	tINIT dancer '=' '{' {for(int j=0; j<6; j++) doublelist[j]=0.0; } inits '}'
				{	if ($2 >= routine->nbDancers || $2 <0) {
						DOERROR("bad dancer index in init");
					}
					for(int j=0; j<6; j++)
						routine->dancerInitPos[$2][j] = doublelist[j];
				}

			|	tKEY dancer time '=' '{' {if (parserMode == COMPILE) key = new Key;} keys '}'
				{	if (parserMode == COMPILE) {
						key->dancer = $2;
						key->start = $3.start; key->end = $3.end;
						routine->keys->append(key);
						if ($2 >= routine->nbDancers || $2 <0) {
							DOERROR("bad dancer index in the key");
						}
						if (key->CheckIntegrity() == FALSE) {
							DOERROR("problem in integrity of key");
						}
					}
				}

			|	tROUTINE {if (parserMode == COMPILE) figure = new Figure();} 
				'(' name ')' time '=' '{' map options '}'
				{	if (parserMode == COMPILE) {
						figure->start = $6.start; figure->end = $6.end; 
						routine->figures->append(figure);
						if (figure->CheckIntegrity() == FALSE) {
							DOERROR("problem in integrity of routine");
						}
					}
				}

			|	tNOTES dancer '=' {if (parserMode == COMPILE) notes = new List<Note>;} notes
				{	if (parserMode == COMPILE) {
						if ($2 == -1) {
							$2 = MAX_DANCERS;
						} else if ($2 >= routine->nbDancers || $2 <0) {
							DOERROR("bad dancer index in the key");
						}
						/* notes[i][0] == notes[i][REAL] */
						if (routine->notes[$2][0] != NULL) {
							DOERROR("notes already defined");
							delete notes;
						} else {
							routine->notes[$2][0] = new List<Note>;
							routine->notes[$2][0]->append(notes);
						}
					}
				}

			|	tMOVIE '(' tSTRING ')' '=' '{' tCENTER '=' path tEYE '=' path '}'
				{	if (parserMode == COMPILE) {
						routine->movies->append(new Movie($3,
								(List<Bezier3D> *) $12.path, (List<Bezier3D> *) $9.path));
					} else {
						routine->movies->append(new Movie($3, 0, 0));
					}
				}
			;


//============= DANCERS ==========================================================================

dancers		:	dancers ',' dancerInfo
				{	routine->dancerNames[routine->nbDancers] = $3.name;
					routine->dancerSex[routine->nbDancers] = $3.sex;
					routine->dancerTypes[routine->nbDancers++] = $3.type; }
			|	dancerInfo
				{	routine->dancerNames[routine->nbDancers] = $1.name;
					routine->dancerTypes[routine->nbDancers++] = $1.type; }
			;

dancerInfo	:	tSTRING								{ $$.name = $1; $$.type = 0; $$.sex = MAN; }
			|	'{' tSTRING ',' tSTRING '}'			{ $$.name = $2; $$.type = $4; $$.sex = MAN; }
			|	'{' sex ',' tSTRING '}'				{ $$.name = $4; $$.type = 0; $$.sex = $2; }
			|	'{' sex ',' tSTRING ',' tSTRING '}'	{ $$.name = $4; $$.type = $6; $$.sex = $2; }
			;

sex			:	tMAN						{ $$ = MAN; }
			|	tWOMAN						{ $$ = WOMAN; }
			|	/* empty */					{ DOERROR("bad sex for dancer"); }
			;

//============= TIME RANGE =======================================================================

time		:	'@' '[' tINTEGER ':' tINTEGER ']'
											{ $$.start = $3; $$.end = $5; }
			|	'@' '[' error ']'
											{ DOERROR("bad time range");
											  $$.start = -1; $$.end = -1; }
			|	'@'							{ DOERROR("need a time range");
											  $$.start = -1; $$.end = -1; }
			|	/* empty */					{ DOERROR("need time range for the key"); }
			;

//============= INIT =============================================================================

inits		:	inits init_entry
			|	/* empty */
			;

init_entry	:	tTRACKNAME '=' number
				{	switch ($1) {
						case cogX:
						case cogY:
						case cogZ:
						case cogRX:
						case cogRY:
						case cogRZ:	doublelist[$1-cogX] = $3;
									break;

						default:	DOERROR("bad init track");
					}
				}
			|	tFILENAME {	DOERROR("bad track name for key"); YYABORT; }
			;

//============= KEY ==============================================================================

dancer		:	'(' tINTEGER ')'			{ $$ = $2; }
			|	/* empty */					{ $$ = 0; DOERROR("need dancer"); }
			;

keys		:	keys key_entry
			|	/* empty */
			;

key_entry	:	tTRACKNAME '=' {if (parserMode == COMPILE) chunks = new List<Chunk>;} chunks
				{	if (parserMode == COMPILE) {
						if (key->part[$1] != NULL) {
							DOERROR("track already defined");
							delete chunks;
						} else {
							if (chunks->head() == NULL) {
								chunks->append(new Chunk($4.time,routine->nbBeats,
										$4.value,$4.value, $4.value,$4.value, -1,-1, TRUE));
							}
							key->part[$1] = chunks; 
						}
					}
				}
			|	tFILENAME {	DOERROR("bad track name for key"); YYABORT; }
			;

//------------- chunks ---------------------------------------------------------------------------
chunks		:	point tLINEAR chunks {
					$$ = $1;
					v1 = $1.value;
					v2 = $3.value;
					if (parserMode == COMPILE) {
						chunks->prepend(new Chunk($1.time,$3.time, v1,v2,
										v1+(v2-v1)/3,v1+(v2-v1)*2/3, -1,-1));
					}
				}
			|	point bezier chunks {
					$$ = $1;
					if (parserMode == COMPILE) {
						chunks->prepend(new Chunk($1.time,$3.time, $1.value,$3.value,
												   $2.v[0],$2.v[1],(int) $2.v[2], (int) $2.v[3]));
					}
				}
			|	point { $$ = $1; }
			|	point ',' chunks { $$ = $1; }
			;

bezier		:	tBEZIER '(' number ',' number ',' number ',' number ')'
						{ $$.v = doublelist;
						  $$.v[0]=$3; $$.v[1]=$5; $$.v[2]=$7; $$.v[3]=$9; }
			|	tBEZIER	error				{ DOERROR("bad bezier parameters"); }
			;

point		:	number '@' tINTEGER		{ $$.time = $3; $$.value = $1; }
			|	/* empty */					{ DOERROR("need a point (value@time)"); }
			;

number		:	tDOUBLE					{ $$ = $1; }
			|	tINTEGER				{ $$ = $1; }
			;


//============= NOTES ============================================================================
notes		:	annotation ',' notes {
					$$ = $1;
					if ($1.end < $1.start ||
						$1.start < 0 || $1.start > routine->nbBeats ||
						$1.end < 0 || $1.end > routine->nbBeats) {
						DOERROR("bad time range for note : exceeds routine or end<start.");
					} else if (parserMode == COMPILE) {
						notes->prepend(new Note($1.start, $1.end, $1.string));
					}
				}
			|	annotation {
					$$ = $1;
					if ($1.end < $1.start ||
						$1.start < 0 || $1.start > routine->nbBeats ||
						$1.end < 0 || $1.end > routine->nbBeats) {
						DOERROR("bat time range for note : exceeds routine or end<start.");
					} else if (parserMode == COMPILE) {
						notes->prepend(new Note($1.start, $1.end, $1.string));
					}
				}
			|	error	{ DOERROR("bad time range for note"); }
			;

annotation	:	tSTRING '@' tINTEGER
				{	$$.start = $$.end = $3;
					if (parserMode == COMPILE) {
						$$.string = $1;
					} else {
						delete[] $1;
					}
				}
			|	tSTRING '@' '[' tINTEGER ':' tINTEGER ']'
				{	$$.start = $4;
					$$.end = $6;
					if (parserMode == COMPILE) {
						$$.string = $1;
					} else {
						delete[] $1;
					}
				}
			|	/* empty */		{ DOERROR("need a note (string@time or string@[time:time])"); }
			;

//============= ROUTINE ==========================================================================

name 		:	tFILENAME					
				{	if (parserMode == COMPILE) {
						figure->filename = $1;
					} else
						delete $1;
				}
			|	/* empty */					{ DOERROR("need a name for the routine"); }

map			:	tMAP '(' map_args ')'
			|	/* empty */					{ DOERROR("need a map statement in the routine"); }
			;

map_args	:	map_args ',' tINTEGER		{ if (parserMode == COMPILE) {
												if (figure->nbMap<MAX_DANCERS)
													figure->map[figure->nbMap++]=$3;
												  if ($3 >= routine->nbDancers || $3 <-1)
													DOERROR("bad dancer index in map");
											  }
											}
			|	tINTEGER 					{ if (parserMode == COMPILE) {
												figure->map[figure->nbMap++]=$1; 
												  if ($1 >= routine->nbDancers || $1 <-1)
													DOERROR("bad dancer index in map");
											  }
											}
			|	/* empty */					{ DOERROR("bad dancer index in map"); }
			;

options		:	options option
			|	/* empty */
			;

option		:	tWITH '{' modifs '}'
			|	tREJECT '{' joints '}'
			|	tCOLOR '=' tSTRING	{	if (parserMode == COMPILE) {
											if (figure->color != NULL) {
												DOERROR("bad dancer index for the modification");
												delete[] $3;
											} else {
												figure->color = $3;
											}
										} else {
											delete[] $3;
										}
									}
			|	tINCLUDENOTES	{  if (parserMode == COMPILE) figure->includeNotes = 1; }
			;

//------------- modifs ---------------------------------------------------------------------------
modifs		:	modifs modif_entry
			|	modif_entry
			;

modif_entry	:	tTRACKNAME dancer '=' {if (parserMode == COMPILE) modifs = new List<Modif>;} modif
				{	if (parserMode == COMPILE) {
						if ($2 >= routine->nbDancers || $2 <-1) {
							DOERROR("bad dancer index for the modification");
							delete modifs;
						} else if (figure->modifications[$2][$1] != NULL) {
							DOERROR("modification already defined for this track");
							delete modifs;
						} else
							figure->modifications[$2][$1] = modifs; 
					}
				}
			|	tFILENAME { DOERROR("bad track name for modification"); }
			;

modif		:	modif ',' action	{ modif.type = 0; }
			|	action				{ modif.type = 0; }
			;

action		:	settings '@' tINTEGER
				{	if (parserMode == COMPILE) {
						modif.time = $3;
						modifs->append(new Modif(&modif));
					}
				}
			|	/* empty */		{ DOERROR("need a action list (list@time)"); }
			;

settings	:	settings '+' setting
			|	setting
			;

setting		:	number
				{	modif.type |= VAL_VALUE;
					modif.data.value = $1;
				}
			|	tHANDLELEFT '(' number ',' number ')'
				{	modif.type |= VAL_LEFT;
					modif.data.leftValue = $3;
					modif.data.leftTime = (int) $5;
				}
			|	tHANDLERIGHT '(' number ',' number ')'
				{	modif.type |= VAL_RIGHT;
					modif.data.rightValue = $3;
					modif.data.rightTime = (int) $5;
				}
			|	tHANDLES '(' number ',' number ',' number ',' number ')'
				{	modif.type |= VAL_LEFT;
					modif.type |= VAL_RIGHT;
					modif.data.leftValue = $3;
					modif.data.leftTime = (int) $5;
					modif.data.rightValue = $7;
					modif.data.rightTime = (int) $9;
				}
			|	/* empty */		{ DOERROR("need a action"); }
			;


//------------- joints ---------------------------------------------------------------------------
joints		:	joints tJOINTNAME dancer
				{	if ($3 >= routine->nbDancers || $3 <0) {
						DOERROR("bad dancer index for the reject statement")
					} else if (parserMode == COMPILE) {
						extern int jointToTrack[][4];
						for(int j=0; j<4 && jointToTrack[$2][j]!=-1; j++)
							figure->flags[$3][jointToTrack[$2][j]] |= REJECT;
					}
				}
			|	tJOINTNAME dancer
				{	if ($2 >= routine->nbDancers || $2 <0) {
						DOERROR("bad dancer index for the reject statement")
					} else if (parserMode == COMPILE) {
						extern int jointToTrack[][4];
						for(int j=0; j<4 && jointToTrack[$2][j]!=-1; j++)
							figure->flags[$2][jointToTrack[$1][j]] |= REJECT;
					}
				}
			|	/* empty */		{ DOERROR("bad track name in 'reject'"); }
			;

//============= PATH (for movie) =================================================================
path		:	{if (parserMode == COMPILE) path = new List<Bezier3D>;} path_entries
				{
					if (parserMode == COMPILE) {
						if (path->head() == NULL) {
							DOERROR("bad path : must contain at least two points");
							delete path;
							path = 0;
						}
						$$.path = path;
					}
				}
			;

path_entries:	point3D '@' tINTEGER bezier3D path_entries
				{	$$ = $1;
					$$.time = $3;
					if (parserMode == COMPILE) {
						class Bezier3D *b = new Bezier3D($3, $5.time);
						b->startX = $1.x; b->startY = $1.y; b->startZ = $1.z;
						b->leftX = $4.v[0]; b->leftY = $4.v[1]; b->leftZ = $4.v[2];
						b->rightX = $4.v[3]; b->rightY = $4.v[4]; b->rightZ = $4.v[5];
						b->endX = $5.x; b->endY = $5.y; b->endZ = $5.z;
						delete[] $4.v;
						path->prepend(b);
					}
				}
			|	point3D '@' tINTEGER
				{	$$ = $1;
					$$.time = $3;
				}
			;

point3D		:	'<' number ',' number ',' number '>' { $$.x = $2; $$.y = $4; $$.z = $6; }
			|	/* empty */					{ DOERROR("need a point3D (<x,y,z>)"); }

bezier3D	:	tBEZIER '(' point3D ',' point3D ')'
				{	$$.v = new double[6];
					$$.v[0]=$3.x; $$.v[1]=$3.y; $$.v[2]=$3.z;
					$$.v[3]=$5.x; $$.v[4]=$5.y; $$.v[5]=$5.z;
				}


%% //=============================================================================================


void yyerror(const char *msg)
{
static char error[100];
extern struct Tcl_Interp *currentInterp;

	sprintf(error, "ERROR: line %d = %s", yylineno, msg);
	Tcl_AppendResult(currentInterp, error, NULL);
}
