#-------------------------------------------------------------------------------------------------
# Commands for management of toolbars
#
#	toolbar::new pathname ?options?					Options passed to the frame
#	toolbar::configure pathname ?options?			Configure main frame containing buttons
#													(e.g. enable/disable)
#
#	toolbar::add:button pathname nameOfButton options
#	toolbar::add:check  pathname nameOfButton options
#	toolbar::add:radio  pathname nameOfButton options
#	toolbar::add:separator ?width?
#
#	toolbar::isOn  pathname nameOfButton
#	toolbar::isOff pathname nameOfButton
#
#	toolbar::check:setOn  pathname nameOfButton
#	toolbar::check:setOff pathname nameOfButton
#	toolbar::check:invoke pathname nameOfButton
#
#	toolbar::radio:setOn  pathname nameOfButton		(not fully operational)
#
#	toolbar::horizontal pathname
#	toolbar::vertical	pathname
#
#-------------------------------------------------------------------------------------------------
#
# BUTTON options :
#		-tip				text to display when user stays on button
#		...					all options of Tk button
#
# CHECK options :
#		-oncommand			command to execute when check becomes On
#		-offcommand			command to execute when check becomes Off
#		-variable			variable to track the state (0 for Off and 1 for On)
#		-tip				text to display when user stays on button
#		... 				all options of Tk button
#
# RADIO options :
#		-oncommand			command to excute (args = nameOfRadio)
#		-offcommand			command to excute (args = nameOfRadio)
#		-group				name of the group to which radio belong
#							(only one radio active in each group)
#		-variable			variable to track the state (name of the active radio by default)
#		-value				value for variable
#		-tip				text to display when user stays on button
#		... 				all options of Tk button
#
#-------------------------------------------------------------------------------------------------

namespace eval toolbar {
}

proc toolbar::new {w args} {
variable $w
upvar 0 $w data

	set data(w:main) [eval frame $w $args]
	set data(v:separator) 0
	set data(v:active) ""
	return $w
}

#-------------------------------------------------------------------------------------------------
# Add new elements : button, check or group of radio
#-------------------------------------------------------------------------------------------------

proc toolbar::add:button {w name args} {
variable $w
upvar 0 $w data

	if {[winfo exists $data(w:main).$name]} {
		error "name $name already used in toolbar $w"
	}

	set tip ""
	foreach {option value} $args {
		switch -exact -- $option {
			-tip		{ set tip $value }
			default		{ lappend opts $option $value }
		}
	}

	pack [eval button $data(w:main).$name $opts] -side left

	if {$tip != ""} {
		registerSlowTip $tip $data(w:main).$name
	}
}

#-------------------------------------------------------------------------------------------------

proc toolbar::add:check {w name args} {
variable $w
upvar 0 $w data
	
	if {[winfo exists $data(w:main).$name]} {
		error "name $name already used in toolbar $w"
	}

	set tip ""
	foreach {option value} $args {
		switch -exact -- $option {
			-oncommand	{ set data(opt:$name:oncommand) $value }
			-offcommand	{ set data(opt:$name:offcommand) $value }
			-variable	{ set data(opt:$name:var) $value }
			-tip		{ set tip $value }
			default		{ lappend opts $option $value }
		}
	}

	set b [eval button $data(w:main).$name $opts -relief raise]
	bindtags $b "$b ToolBarCheck all"
	pack $b -side left

	if {[info exists data(opt:$name:var)]} {
		trace vdelete ::$data(opt:$name:var) w "toolbar::check:toggle $w $b"
		trace variable ::$data(opt:$name:var) w "toolbar::check:toggle $w $b"
	}
	if {$tip != ""} {
		registerSlowTip $tip $data(w:main).$name
	}
	set data(t:done) 0
}

#-------------------------------------------------------------------------------------------------

proc toolbar::add:radio {w name args} {
variable $w
upvar 0 $w data

	
	if {[winfo exists $data(w:main).$name]} {
		error "name $name already used in toolbar $w"
	}

	set tip ""
	set group 0
	foreach {option value} $args {
		switch -exact -- $option {
			-oncommand	{ set data(opt:$name:oncommand) $value }
			-offcommand	{ set data(opt:$name:offcommand) $value }
			-group		{ set group $value }
			-variable	{ set data(opt:$name:var) $value }
			-value		{ set data(opt:$name:value) $value }
			-tip		{ set tip $value }
			default		{ lappend opts $option $value }
		}
	}

	set b [eval button $data(w:main).$name $opts -relief raise]
	bindtags $b "$b ToolBarRadio all"
	pack $b -side left

	lappend data(v:group:$group) $b
	set data(opt:$name:group) $group

	if {[info exists data(opt:$name:var)]} {
		trace vdelete ::$data(opt:$name:var) w "toolbar::radio:toggle $w $group"
		trace variable ::$data(opt:$name:var) w "toolbar::radio:toggle $w $group"
	}
	if {$tip != ""} {
		registerSlowTip $tip $data(w:main).$name
	}
}

#-------------------------------------------------------------------------------------------------

proc toolbar::add:separator {w {size 3}} {
variable $w
upvar 0 $w data

	
	frame $data(w:main).$data(v:separator) -width $size -height $size
	pack $data(w:main).$data(v:separator) -side left
	incr data(v:separator)
}

#=================================================================================================

proc toolbar::isOn {w name} {
variable $w
upvar 0 $w data

	if {[$data(w:main).$name cget -relief] == "sunken"} {
		return 1
	} else {
		return 0
	}
}

proc toolbar::isOff {w name} {
variable $w
upvar 0 $w data

	if {[$data(w:main).$name cget -relief] == "raised"} {
		return 1
	} else {
		return 0
	}
}

#-------------------------------------------------------------------------------------------------

proc toolbar::configure {w name args} {
variable $w
upvar 0 $w data

	eval $data(w:main).$name configure $args
}

#-------------------------------------------------------------------------------------------------

proc toolbar::vertical {w} {
variable $w
upvar 0 $w data


	foreach w [pack slaves $data(w:main)] {
		pack configure $w -side top
	}
}

proc toolbar::horizontal {w} {
variable $w
upvar 0 $w data


	foreach w [pack slaves $data(w:main)] {
		pack configure $w -side left
	}
}

#=================================================================================================

bind ToolBarCheck <ButtonPress-1> "toolbar::check:press %W"
bind ToolBarCheck <ButtonRelease-1> "toolbar::check:release %W"
bind ToolBarCheck <Enter> "toolbar::check:enter %W"
bind ToolBarCheck <Leave> "toolbar::check:leave %W"

proc toolbar::check:enter {w} {
variable [winfo parent $w]
upvar 0 [winfo parent $w] data

	if {[$w cget -state] != "disabled"} {
		$w configure -state active
		if {$data(v:active) == $w} {
			$w configure -relief $data(v:enter)
		}
		set data(v:window) $w
	}
}

proc toolbar::check:leave {w} {
variable [winfo parent $w]
upvar 0 [winfo parent $w] data

	if {[$w cget -state] != "disabled"} {
		$w configure -state normal
		if {$data(v:active) == $w} {
			$w configure -relief $data(v:leave)
		}
	}
}

proc toolbar::check:press {w} {
variable [winfo parent $w]
upvar 0 [winfo parent $w] data

	if {[$w cget -state] != "disabled"} {
		set data(v:active) $w
		if {[$w cget -relief] == "sunken"} {
			set data(v:enter) "raised"
			set data(v:leave) "sunken"
			$w configure -state active -relief $data(v:enter)
		} else {
			set data(v:enter) "sunken"
			set data(v:leave) "raised"
			$w configure -state active -relief $data(v:enter)
		}
	}
}

proc toolbar::check:release {w} {
variable [winfo parent $w]
upvar 0 [winfo parent $w] data


	set data(v:active) ""
	if {[$w cget -state] != "disabled" && $data(v:window) == $w} {
		$w configure -state normal
		set window [winfo name $w]
		if {[$w cget -relief] == "sunken" && [info exists data(opt:$window:oncommand)]} {
			uplevel #0 $data(opt:$window:oncommand)
		} elseif {[$w cget -relief] == "raised" && [info exists data(opt:$window:offcommand)]} {
			uplevel #0 $data(opt:$window:offcommand)
		}
		# already called fn : no call in toggle
		set data(t:done) 1

		# if variable defined
		if {[info exists data(opt:$window:var)]} {
			if {[$w cget -relief] == "sunken"} {
				uplevel #0 "set ::$data(opt:$window:var) 1"
			} else {
				uplevel #0 "set ::$data(opt:$window:var) 0"
			}
		}
	}
}

proc toolbar::check:toggle {w widget name1 name2 op} {
variable $w
upvar 0 $w data

	if {$name2 == ""} {
		set value [uplevel #0 set ::$name1]
	} else {
		set name ${name1}($name2)
		set value [uplevel #0 set ::$name]
	}

	set window [winfo name $widget]
	if {$value} {
		$widget configure -relief sunken
		if {!$data(t:done) && [info exists data(opt:$window:oncommand)]} {
			uplevel #0 $data(opt:$window:oncommand)
		}
	} elseif {!$data(t:done)} {
		$widget configure -relief raised
		if {[info exists data(opt:$window:offcommand)]} {
			uplevel #0 $data(opt:$window:offcommand)
		}
	}
	set data(t:done) 0
}

proc toolbar::check:setOn {w name} {
variable $w
upvar 0 $w data


	$data(w:main).$name configure -relief sunken

	# if oncommand defind
	if {[info exists data(opt:$name:oncommand)]} {
		uplevel #0 $data(opt:$name:oncommand)
	}
	# if variable defined
	if {[info exists data(opt:$name:var)]} {
		uplevel #0 "set ::$data(opt:$name:var) 1"
	}
}

proc toolbar::check:setOff {w name} {
variable $w
upvar 0 $w data


	$data(w:main).$name configure -relief raised

	# if oncommand defind
	if {[info exists data(opt:$name:offcommand)]} {
		uplevel #0 $data(opt:$name:offcommand)
	}
	# if variable defined
	if {[info exists data(opt:$name:var)]} {
		uplevel #0 "set ::$data(opt:$name:var) 0"
	}
}

proc toolbar::check:invoke {w name} {
variable $w
upvar 0 $w data

	set data(v:window) $w.$name
	check:press $w.$name
	check:release $w.$name
}

#-------------------------------------------------------------------------------------------------

bind ToolBarRadio <ButtonPress-1> "toolbar::radio:press %W"
bind ToolBarRadio <ButtonRelease-1> "toolbar::radio:release %W"
bind ToolBarRadio <Enter> "toolbar::check:enter %W"
bind ToolBarRadio <Leave> "toolbar::check:leave %W"


proc toolbar::radio:press {w} {
variable [winfo parent $w]
upvar 0 [winfo parent $w] data

	if {[$w cget -state] != "disabled" && [$w cget -relief] == "raised"} {
		set data(v:active) $w
		set data(v:enter) "sunken"
		set data(v:leave) "raised"
		$w configure -state active -relief $data(v:enter)
	}
}

proc toolbar::radio:release {w} {
variable [winfo parent $w]
upvar 0 [winfo parent $w] data

	set data(v:active) ""
	if {[$w cget -state] != "disabled" && $data(v:window) == $w} {
		$w configure -state normal -relief sunken
		set window [winfo name $w]
		if {[$w cget -relief] == "sunken" && [info exists data(opt:$window:oncommand)]} {
			uplevel #0 $data(opt:$window:oncommand) $window
		} elseif {[info exists data(opt:$window:offcommand)]} {
			uplevel #0 $data(opt:$window:offcommand) $window
		}
		# update other buttons in same group
		foreach b $data(v:group:$data(opt:$window:group)) {
			if {$b != $w} {
				if {[info exists data(opt:$window:offcommand)]} {
					uplevel #0 $data(opt:[winfo name $b]:offcommand) [winfo name $b]
				}
				$b configure -relief raised
			}
		}

		# if variable defined
		if {[info exists data(opt:$window:var)]} {
			if {[info exists data(opt:$window:value)]} {
				uplevel #0 "set ::$data(opt:$window:var) $data(opt:$window:value)"
			} else {
				uplevel #0 "set ::$data(opt:$window:var) $window"
			}
		}
	}
}

proc toolbar::radio:toggle {w group name1 name2 op} {
variable $w
upvar 0 $w data


	if {$name2 == ""} {
		set value [uplevel #0 set ::$name1]
	} else {
		set name ${name1}($name2)
		set value [uplevel #0 set ::$name]
	}

	foreach b $data(v:group:$group) {
		if {[winfo name $b] == $value} {
			set data(v:window) $b
			break;
		}
		if {[info exists data(opt:[winfo name $b]:value)]} {
			if {$data(opt:[winfo name $b]:value) == $value} {
				set data(v:window) $b
				break;
			}
		}
	}

	radio:release $data(v:window)
}

proc toolbar::radio:setOn {w name} {
variable $w
upvar 0 $w data

	$data(w:main).$name configure -relief sunken
}


##################################################################################################


proc dummy {} {
global var
global var2

	foreach w [winfo children .] { destroy $w }

	canvas .c -bg red
	pack .c

	catch {unset toolbar::.t}

	toolbar::new .t -bg green
	pack .t -expand yes -anchor w

	toolbar::add:button .t toto -bd 1 -text "toto" -command "puts coucou"
	toolbar::add:button .t titi -bd 1 -text "titi" -command "puts cici"

	toolbar::add:separator .t

	toolbar::add:check .t tata -bd 1 -text "tata" -oncommand "puts on"
	toolbar::add:check .t tutu -bd 1 -text "tutu" -offcommand "puts off"
	toolbar::add:check .t tutu2 -bd 1 -text "tutu var" -variable var2

	toolbar::add:separator .t

	toolbar::add:radio .t ra1 -bd 1 -text "ra1" -group a -variable var
	toolbar::add:radio .t ra2 -bd 1 -text "ra2" -group a -variable var

	toolbar::add:separator .t

	toolbar::add:radio .t rb1 -bd 1 -text "rb1" -group b -oncommand "test"
	toolbar::add:radio .t rb2 -bd 1 -text "rb2" -group b -offcommand "test"
	toolbar::add:radio .t rb3 -bd 1 -text "rb3" -group b -oncommand "test"


	set var ra1

	proc check {w mode} {
		puts "CHECK  $w , $mode"
	}

	proc test {w mode} {
		puts "RADIO  $w , $mode"
	}
}
