#ifndef INCLUDED_VIEW
#define INCLUDED_VIEW

#include "global.h"
#include "list.h"

class Body;
class Bezier;

//----------------------------------------------------------------------------------------------- 
//	Floor
//----------------------------------------------------------------------------------------------- 

struct Floor {
	int nbX, nbZ;
	double padX, padZ;
	double color[4], colorAxe[4];

	Floor()	{ nbX = nbZ = 10;
			  padX = padZ = 0.5;
			  color[0] = color[1] = color[2] = .8; color[3] = 0.0;
			  colorAxe[0] = .9; colorAxe[1] = 1.0; colorAxe[2] = .75; colorAxe[3] = 0.0; }

	void Draw() const;
};

//----------------------------------------------------------------------------------------------- 
//	Viewpoint
//----------------------------------------------------------------------------------------------- 

struct Viewpoint {
	double azimut, elevation, twist, dist;
	double xTrans, yTrans;
	double xPos, yPos, zPos;
	double fovy, zNear, zFar;
	double projScale;

	Viewpoint()					{ azimut = elevation = twist = 0.0; dist = 6.0;
								  xTrans = yTrans = 0.0;
								  xPos = yPos = zPos = 0.0;
								  fovy = 30.0; zNear = 0.05; zFar = 100.0;
								}
};

//----------------------------------------------------------------------------------------------- 
//	Light
//----------------------------------------------------------------------------------------------- 

class Light {
	// DATA
	int number;
	char *name;

	double position[4];
	double ambient[4];
	double diffuse[4];

	// FRIENDS
	friend class View3D;
	friend class ViewMovie;
	friend int Light_ObjCmd(ClientData clientData, Tcl_Interp *interp,
						    int objc, Tcl_Obj *CONST objv[]);

public:
	static class List<Togl> *views;
	static int changed;

	// CREATORS
	Light(int counter, char *n);
	~Light() { delete[] name; }

	// MANIPULATORS
	int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

	// ACCESSORS
	int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
};


//================================================================================================
//
//		Views (rendered via OpenGL)
//
//================================================================================================

class AbstractView {
public:
	AbstractView() {}
	virtual ~AbstractView() {}
	virtual void Redisplay() const = 0;
private:
	virtual void SetGLView(int picking, int x, int y) const = 0;
};

class ViewMovie;

//------------------------------------------------------------------------------------------------
//	3D view
//------------------------------------------------------------------------------------------------

struct LocalAxis {
	double X, Y, Z, pad, color1[4], color2[4];

	LocalAxis() { color1[0] = color1[1] = color1[2] = 1.0; color1[3] = 0.0;
				  color2[0] = color2[1] = color2[2] = 1.0; color2[3] = 0.0; }
};

class View3D : public AbstractView {
	// DATA
	struct Togl *togl;
	class Viewpoint viewpoint;
	class Floor floor;
	class List<Body> *bodies;

	int forceWire, jointsOnly, withRX, withRY, withRZ, setupLight;
	int drawFloor, drawAxis, drawLocalAxis, sphere, sphereX, sphereZ, disk;
	double size, matrix[16];
	struct LocalAxis localAxis;

	friend ViewMovie;

public:
	int ortho;
	static class List<Light> *lights;

	// CREATORS
	View3D(struct Togl *t);
	~View3D();

	// MANIPULATORS
	int Add(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Remove(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

	// ACCESSORS
	int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Pick(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int joint) const;
	void Redisplay() const;
	int Snapshot(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]); // no const because 'togl'
																	   // modifed for shrink

private:
	virtual void SetGLView(int picking, int x, int y) const;
};

//------------------------------------------------------------------------------------------------
//	Movie view
//------------------------------------------------------------------------------------------------

class Bezier3D;

class ViewMovie : public View3D {
	// DATA
	enum {EDIT, VIEW, PLAY};
	int mode, showHeight, showHandle, showLine, showPaths, setupLight;
	int currentTime, askedTime;
	double colorEye[4], colorCenter[4];

	class List<Bezier3D> *eye, *center;
	class ListIterator<Bezier3D> currentEye, currentCenter;
	//class List<Bezier> *twist;
	//class ListIterator<Bezier> currentTwist;

	static GLuint Font;

	// CREATORS
public:
	ViewMovie(struct Togl *t);
	~ViewMovie();

	// MANIPULATORS
	int Modify(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Record(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Viewpoint(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

	// ACCESSORS
	int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Info(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Pick(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	void Redisplay() const;

private:
	virtual void SetGLView(int picking, int x, int y) const;
	void SetTime(int time);
	enum {SET_POINT, MOVE_POINT, LEFT_HANDLE, RIGHT_HANDLE, SCALE_HANDLE, HANDLE_TO_POINT};
	void ApplyModif(List<Bezier3D> *path, int time, int type, double X, double Y, double Z);
	void AlignHandles(List<Bezier3D> *path, int time);
	void DeletePoint(List<Bezier3D> *path, int time);
	void Insert(List<Bezier3D> *path, double X, double Y, double Z);
};


//----------------------------------------------------------------------------------------------- 
//	Track view (spline)
//----------------------------------------------------------------------------------------------- 

struct Color {
	double bg[4], grid[4], time[4], limits[4], zoomBox[4];
	double point[4], line[MAX_VIEWTRACK][4], handle[4], selected[4];

	Color() 	{ bg[0] = bg[1] = bg[2] = bg[3] = 0.0;
				  grid[0] = grid[1] = grid[2] = 0.6; grid[3] = 0.0;
				  time[0] = time[1] = time[2] = 0.3; time[3] = 0.0;
				  limits[0] = limits[1] = 1.0; limits[2] = limits[3] = 0.0;
				  zoomBox[0] = zoomBox[1] = zoomBox[2] = zoomBox[3] = 0.0;
				  point[0] = 1.0; point[1] = point[2] = point[3] = 0.0;
				  for (int i=0; i<MAX_VIEWTRACK; i++) {
					  line[i][0] = line[i][1] = line[i][2] = 1.0; line[i][3] = 0.0;
				  }
				  handle[0] = handle[2] = handle[3] = 0.0; handle[1] = 1.0;
				  selected[0] = selected[1] = 1.0; selected[2] = selected[3] = 0.0;
				}
};


class ViewTrack : public AbstractView {
	// DATA
	struct Togl *togl;

	int maxTime, maxY;					// to define area to draw
	double xPos, yPos, xZoom, yZoom;	// position in the view
	int xmin, xmax; 					// min and max coord of the view & pos for scrolling
	double ymin, ymax;

	int zbXmin, zbXmax;			// internal vars to memorize the Zoom Box
	double zbYmin, zbYmax;
	int stippleZoomBox;			// stipple pattern for the zoom box

	int limitMin, limitMax;		// to draw the limits of a joint
	int currentTime;			// to draw the line representing the current time
	int padTime, stippleTime, padTime2, stippleTime2;	// to draw the grid (Time)
	int padY, stippleY, padY2, stippleY2;				// to draw the grid (Y axis)

	double pointSize;			// point size for the handle

	int drawGrid, drawCurrentTime, drawRoutines, drawHandles;	// flags

	struct Color color;			// colors used in the drawing

	class Routine *routine;		// routine used for the view
	int dancer;					// dancer index in this routine
	int track[MAX_VIEWTRACK];	// indices of track to represent

	enum {VT_LINE, VT_POINT, VT_HANDLE};

public:
	// CREATORS
	ViewTrack(struct Togl *t);
	~ViewTrack();

	// MANIPULATORS
	int Convert(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Insert(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Limits(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Modify(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Move(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Zoom(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

	// ACCESSORS
	void Draw(int draw_mode, int i) const;
	int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Pick(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int PickFigure(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int askID) const;
	void Redisplay() const;

private:
	virtual void SetGLView(int picking, int x, int y) const;
	void ScreenToView(int x, int y, int *vx, double *vy);
};


#endif
